print("import: 'fuel'")
import fuel

print("import: 'fuel.bin'")
import fuel.bin

print("import: 'fuel.converters'")
import fuel.converters

print("import: 'fuel.datasets'")
import fuel.datasets

print("import: 'fuel.downloaders'")
import fuel.downloaders

print("import: 'fuel.transformers'")
import fuel.transformers

print("import: 'fuel.utils'")
import fuel.utils

